@@ -278,3 +278,24 @@ INSTANTIATE_TEST_SUITE_P(len1_swap_2D,
                              false,
                              false)),
                          accuracy_test::TestName);
+
+const static std::vector<size_t> lengths_for_disabled_autoalloc
+    = merge_and_sort_values<size_t>({pow2_range, pow3_range, prime_range, mix_range}, 12);
+
+INSTANTIATE_TEST_SUITE_P(
+    various_2D,
+    accuracy_test,
+    ::testing::ValuesIn(param_generator(test_prob,
+                                        generate_lengths({lengths_for_disabled_autoalloc,
+                                                          lengths_for_disabled_autoalloc}),
+                                        precision_range_sp_dp,
+                                        batch_range,
+                                        stride_range,
+                                        stride_range,
+                                        ioffset_range_zero,
+                                        ooffset_range_zero,
+                                        place_range,
+                                        false,
+                                        false,
+                                        fft_auto_allocation_off)),
+    accuracy_test::TestName);
