@@ -19,6 +19,7 @@
 // THE SOFTWARE.
 
 #include <boost/scope_exit.hpp>
+#include <boost/tokenizer.hpp>
 #include <gtest/gtest.h>
 #include <hip/hip_runtime.h>
 #include <math.h>
@@ -33,15 +34,36 @@
 #include "../../shared/accuracy_test.h"
 #include "../../shared/fftw_transform.h"
 #include "../../shared/gpubuf.h"
+#include "../../shared/params_gen.h"
 #include "../../shared/rocfft_against_fftw.h"
 #include "../../shared/rocfft_complex.h"
+#include "../../shared/subprocess.h"
+
+extern std::string mp_launch;
+
+extern last_cpu_fft_cache last_cpu_fft_data;
+
+// clang-format off
+// tokens of tests found to be symptomatic
+static const std::vector<std::string> symptomatic_tokens = {
+#ifndef _CUFFT_BACKEND
+// cases specific to ROCM backend
+#else
+    // cases specific to CUFFT backend
+    "real_forward_len_16384_half_ip_batch_4_istride_1_R_ostride_1_HI_idist_16386_odist_8193_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_32768_half_ip_batch_4_istride_1_R_ostride_1_HI_idist_32770_odist_16385_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_65536_half_ip_batch_2_istride_1_R_ostride_1_HI_idist_65538_odist_32769_ioffset_0_0_ooffset_0_0",
+#endif
+    // common  to both backends
+};
+// clang-format on
 
 void fft_vs_reference(hipfft_params& params, bool round_trip)
 {
     switch(params.precision)
     {
     case fft_precision_half:
-        fft_vs_reference_impl<_Float16, hipfft_params>(params, round_trip);
+        fft_vs_reference_impl<rocfft_fp16, hipfft_params>(params, round_trip);
         break;
     case fft_precision_single:
         fft_vs_reference_impl<float, hipfft_params>(params, round_trip);
@@ -68,28 +90,91 @@ TEST_P(accuracy_test, vs_fftw)
         GTEST_SKIP();
     }
 
-    // only do round trip for single-GPU FFTs
-    bool round_trip = params.multiGPU <= 1;
-
-    if(!params.run_callbacks)
+    switch(params.mp_lib)
     {
+    case fft_params::fft_mp_lib_none:
+    {
+        // skipping symptomatic case(s), unless forcefully/knowingly executing normally-disabled
+        // test tokens (e.g., by using --gtest_also_run_disabled_tests)
+        const char* test_suite_name
+            = ::testing::UnitTest::GetInstance()->current_test_info()->test_suite_name();
+        if(!symptomatic_tokens.empty() && std::strstr(test_suite_name, "DISABLED") == nullptr
+           && std::find(symptomatic_tokens.begin(), symptomatic_tokens.end(), params.token())
+                  != symptomatic_tokens.end())
+        {
+            GTEST_SKIP()
+                << "Symptomatic test that's currently disabled by default (force-skipping). Use "
+                   "CLI arguments '--gtest_also_run_disabled_tests' to force the test execution "
+                   "(via another test suite).";
+        }
+        // only do round trip for forward FFTs
+        const bool do_round_trip = params.is_forward();
+
         try
         {
-            fft_vs_reference(params, round_trip);
+            fft_vs_reference(params, do_round_trip);
         }
-        catch(ROCFFT_GTEST_SKIP& e)
+        catch(HOSTBUF_MEM_USAGE& e)
         {
-            GTEST_SKIP() << e.msg.str();
+            // explicitly clear cache
+            last_cpu_fft_data = last_cpu_fft_cache();
+            GTEST_SKIP() << e.msg;
         }
-        catch(ROCFFT_GTEST_FAIL& e)
+        catch(ROCFFT_SKIP& e)
         {
-            GTEST_FAIL() << e.msg.str();
+            GTEST_SKIP() << e.msg;
         }
+        catch(const fft_params::unimplemented_exception& e)
+        {
+            GTEST_SKIP() << "Unimplemented exception: " << e.what();
+        }
+        catch(ROCFFT_FAIL& e)
+        {
+            GTEST_FAIL() << e.msg;
+        }
+        break;
+    }
+    case fft_params::fft_mp_lib_mpi:
+    {
+        // split launcher into tokens since the first one is the exe
+        // and the remainder is the start of its argv
+        boost::escaped_list_separator<char>                   sep('\\', ' ', '\"');
+        boost::tokenizer<boost::escaped_list_separator<char>> tokenizer(mp_launch, sep);
+        std::string                                           exe;
+        std::vector<std::string>                              argv;
+        for(auto t : tokenizer)
+        {
+            if(t.empty())
+                continue;
+
+            if(exe.empty())
+                exe = t;
+            else
+                argv.push_back(t);
+        }
+        // append test token and ask for accuracy test
+        argv.push_back("--token");
+        argv.push_back(params.token());
+        argv.push_back("--accuracy");
+
+        // throws an exception if launch fails or if subprocess
+        // returns nonzero exit code
+        execute_subprocess(exe, argv, {});
+        break;
+    }
+    default:
+        GTEST_FAIL() << "Invalid communicator choice!";
+        break;
     }
 
     SUCCEED();
 }
 
+INSTANTIATE_TEST_SUITE_P(DISABLED_symptomatic_tokens,
+                         accuracy_test,
+                         ::testing::ValuesIn(param_generator_token(test_prob, symptomatic_tokens)),
+                         accuracy_test::TestName);
+
 #ifdef __HIP__
 
 // load/store callbacks - cbdata in each is actually a scalar double
@@ -109,13 +194,44 @@ __host__ __device__ Tdata load_callback(Tdata* input, size_t offset, void* cbdat
     }
 }
 
-__device__ auto load_callback_dev_half           = load_callback<_Float16>;
-__device__ auto load_callback_dev_complex_half   = load_callback<rocfft_complex<_Float16>>;
+__device__ auto load_callback_dev_half           = load_callback<rocfft_fp16>;
+__device__ auto load_callback_dev_complex_half   = load_callback<rocfft_complex<rocfft_fp16>>;
 __device__ auto load_callback_dev_float          = load_callback<float>;
 __device__ auto load_callback_dev_complex_float  = load_callback<rocfft_complex<float>>;
 __device__ auto load_callback_dev_double         = load_callback<double>;
 __device__ auto load_callback_dev_complex_double = load_callback<rocfft_complex<double>>;
 
+// load/store callbacks - cbdata in each is actually a scalar double
+// with a number to apply to each element
+template <typename Tdata>
+__host__ __device__ Tdata
+    load_callback_round_trip_inverse(Tdata* input, size_t offset, void* cbdata, void* sharedMem)
+{
+    auto testdata = static_cast<const callback_test_data*>(cbdata);
+    // subtract each element by scalar
+    if(input == testdata->base)
+        return input[offset] - testdata->scalar;
+    // wrong base address passed, return something obviously wrong
+    else
+    {
+        // wrong base address passed, return something obviously wrong
+        return input[0];
+    }
+}
+
+__device__ auto load_callback_round_trip_inverse_dev_half
+    = load_callback_round_trip_inverse<rocfft_fp16>;
+__device__ auto load_callback_round_trip_inverse_dev_complex_half
+    = load_callback_round_trip_inverse<rocfft_complex<rocfft_fp16>>;
+__device__ auto load_callback_round_trip_inverse_dev_float
+    = load_callback_round_trip_inverse<float>;
+__device__ auto load_callback_round_trip_inverse_dev_complex_float
+    = load_callback_round_trip_inverse<rocfft_complex<float>>;
+__device__ auto load_callback_round_trip_inverse_dev_double
+    = load_callback_round_trip_inverse<double>;
+__device__ auto load_callback_round_trip_inverse_dev_complex_double
+    = load_callback_round_trip_inverse<rocfft_complex<double>>;
+
 void* get_load_callback_host(fft_array_type itype,
                              fft_precision  precision,
                              bool           round_trip_inverse = false)
@@ -129,22 +245,55 @@ void* get_load_callback_host(fft_array_type itype,
         switch(precision)
         {
         case fft_precision_half:
-            EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
-                                          HIP_SYMBOL(load_callback_dev_complex_half),
-                                          sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(
+                              &load_callback_host,
+                              HIP_SYMBOL(load_callback_round_trip_inverse_dev_complex_half),
+                              sizeof(void*)),
+                          hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
+                                              HIP_SYMBOL(load_callback_dev_complex_half),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return load_callback_host;
         case fft_precision_single:
-            EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
-                                          HIP_SYMBOL(load_callback_dev_complex_float),
-                                          sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(
+                              &load_callback_host,
+                              HIP_SYMBOL(load_callback_round_trip_inverse_dev_complex_float),
+                              sizeof(void*)),
+                          hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
+                                              HIP_SYMBOL(load_callback_dev_complex_float),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return load_callback_host;
         case fft_precision_double:
-            EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
-                                          HIP_SYMBOL(load_callback_dev_complex_double),
-                                          sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(
+                              &load_callback_host,
+                              HIP_SYMBOL(load_callback_round_trip_inverse_dev_complex_double),
+                              sizeof(void*)),
+                          hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
+                                              HIP_SYMBOL(load_callback_dev_complex_double),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return load_callback_host;
         }
     }
@@ -153,19 +302,55 @@ void* get_load_callback_host(fft_array_type itype,
         switch(precision)
         {
         case fft_precision_half:
-            EXPECT_EQ(hipMemcpyFromSymbol(
-                          &load_callback_host, HIP_SYMBOL(load_callback_dev_half), sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
+                                              HIP_SYMBOL(load_callback_round_trip_inverse_dev_half),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
+                                              HIP_SYMBOL(load_callback_dev_half),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return load_callback_host;
         case fft_precision_single:
-            EXPECT_EQ(hipMemcpyFromSymbol(
-                          &load_callback_host, HIP_SYMBOL(load_callback_dev_float), sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(
+                    hipMemcpyFromSymbol(&load_callback_host,
+                                        HIP_SYMBOL(load_callback_round_trip_inverse_dev_float),
+                                        sizeof(void*)),
+                    hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
+                                              HIP_SYMBOL(load_callback_dev_float),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return load_callback_host;
         case fft_precision_double:
-            EXPECT_EQ(hipMemcpyFromSymbol(
-                          &load_callback_host, HIP_SYMBOL(load_callback_dev_double), sizeof(void*)),
-                      hipSuccess);
+
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(
+                    hipMemcpyFromSymbol(&load_callback_host,
+                                        HIP_SYMBOL(load_callback_round_trip_inverse_dev_double),
+                                        sizeof(void*)),
+                    hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&load_callback_host,
+                                              HIP_SYMBOL(load_callback_dev_double),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return load_callback_host;
         }
     }
@@ -188,13 +373,38 @@ __host__ __device__ static void
     // otherwise, wrong base address passed, just don't write
 }
 
-__device__ auto store_callback_dev_half           = store_callback<_Float16>;
-__device__ auto store_callback_dev_complex_half   = store_callback<rocfft_complex<_Float16>>;
+__device__ auto store_callback_dev_half           = store_callback<rocfft_fp16>;
+__device__ auto store_callback_dev_complex_half   = store_callback<rocfft_complex<rocfft_fp16>>;
 __device__ auto store_callback_dev_float          = store_callback<float>;
 __device__ auto store_callback_dev_complex_float  = store_callback<rocfft_complex<float>>;
 __device__ auto store_callback_dev_double         = store_callback<double>;
 __device__ auto store_callback_dev_complex_double = store_callback<rocfft_complex<double>>;
 
+template <typename Tdata>
+__host__ __device__ static void store_callback_round_trip_inverse(
+    Tdata* output, size_t offset, Tdata element, void* cbdata, void* sharedMem)
+{
+    auto testdata = static_cast<callback_test_data*>(cbdata);
+    // divide each element by scalar
+    if(output == testdata->base)
+    {
+        output[offset] = element / testdata->scalar;
+    }
+    // otherwise, wrong base address passed, just don't write
+}
+__device__ auto store_callback_round_trip_inverse_dev_half
+    = store_callback_round_trip_inverse<rocfft_fp16>;
+__device__ auto store_callback_round_trip_inverse_dev_complex_half
+    = store_callback_round_trip_inverse<rocfft_complex<rocfft_fp16>>;
+__device__ auto store_callback_round_trip_inverse_dev_float
+    = store_callback_round_trip_inverse<float>;
+__device__ auto store_callback_round_trip_inverse_dev_complex_float
+    = store_callback_round_trip_inverse<rocfft_complex<float>>;
+__device__ auto store_callback_round_trip_inverse_dev_double
+    = store_callback_round_trip_inverse<double>;
+__device__ auto store_callback_round_trip_inverse_dev_complex_double
+    = store_callback_round_trip_inverse<rocfft_complex<double>>;
+
 void* get_store_callback_host(fft_array_type otype,
                               fft_precision  precision,
                               bool           round_trip_inverse = false)
@@ -208,22 +418,55 @@ void* get_store_callback_host(fft_array_type otype,
         switch(precision)
         {
         case fft_precision_half:
-            EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
-                                          HIP_SYMBOL(store_callback_dev_complex_half),
-                                          sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(
+                              &store_callback_host,
+                              HIP_SYMBOL(store_callback_round_trip_inverse_dev_complex_half),
+                              sizeof(void*)),
+                          hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
+                                              HIP_SYMBOL(store_callback_dev_complex_half),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return store_callback_host;
         case fft_precision_single:
-            EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
-                                          HIP_SYMBOL(store_callback_dev_complex_float),
-                                          sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(
+                              &store_callback_host,
+                              HIP_SYMBOL(store_callback_round_trip_inverse_dev_complex_float),
+                              sizeof(void*)),
+                          hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
+                                              HIP_SYMBOL(store_callback_dev_complex_float),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return store_callback_host;
         case fft_precision_double:
-            EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
-                                          HIP_SYMBOL(store_callback_dev_complex_double),
-                                          sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(
+                              &store_callback_host,
+                              HIP_SYMBOL(store_callback_round_trip_inverse_dev_complex_double),
+                              sizeof(void*)),
+                          hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
+                                              HIP_SYMBOL(store_callback_dev_complex_double),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return store_callback_host;
         }
     }
@@ -232,21 +475,55 @@ void* get_store_callback_host(fft_array_type otype,
         switch(precision)
         {
         case fft_precision_half:
-            EXPECT_EQ(hipMemcpyFromSymbol(
-                          &store_callback_host, HIP_SYMBOL(store_callback_dev_half), sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(
+                    hipMemcpyFromSymbol(&store_callback_host,
+                                        HIP_SYMBOL(store_callback_round_trip_inverse_dev_half),
+                                        sizeof(void*)),
+                    hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
+                                              HIP_SYMBOL(store_callback_dev_half),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return store_callback_host;
         case fft_precision_single:
-            EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
-                                          HIP_SYMBOL(store_callback_dev_float),
-                                          sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(
+                    hipMemcpyFromSymbol(&store_callback_host,
+                                        HIP_SYMBOL(store_callback_round_trip_inverse_dev_float),
+                                        sizeof(void*)),
+                    hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
+                                              HIP_SYMBOL(store_callback_dev_float),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return store_callback_host;
         case fft_precision_double:
-            EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
-                                          HIP_SYMBOL(store_callback_dev_double),
-                                          sizeof(void*)),
-                      hipSuccess);
+            if(round_trip_inverse)
+            {
+                EXPECT_EQ(
+                    hipMemcpyFromSymbol(&store_callback_host,
+                                        HIP_SYMBOL(store_callback_round_trip_inverse_dev_double),
+                                        sizeof(void*)),
+                    hipSuccess);
+            }
+            else
+            {
+                EXPECT_EQ(hipMemcpyFromSymbol(&store_callback_host,
+                                              HIP_SYMBOL(store_callback_dev_double),
+                                              sizeof(void*)),
+                          hipSuccess);
+            }
             return store_callback_host;
         }
     }
@@ -275,10 +552,11 @@ void apply_store_callback(const fft_params& params, std::vector<hostbuf>& output
         {
         case fft_precision_half:
         {
-            const size_t elem_size = sizeof(std::complex<_Float16>);
+            const size_t elem_size = sizeof(std::complex<rocfft_fp16>);
             const size_t num_elems = output.front().size() / elem_size;
 
-            auto output_begin = reinterpret_cast<rocfft_complex<_Float16>*>(output.front().data());
+            auto output_begin
+                = reinterpret_cast<rocfft_complex<rocfft_fp16>*>(output.front().data());
             for(size_t i = 0; i < num_elems; ++i)
             {
                 auto& element = output_begin[i];
@@ -332,12 +610,12 @@ void apply_store_callback(const fft_params& params, std::vector<hostbuf>& output
         {
         case fft_precision_half:
         {
-            const size_t elem_size = sizeof(std::complex<_Float16>);
+            const size_t elem_size = sizeof(std::complex<rocfft_fp16>);
             for(auto& buf : output)
             {
                 const size_t num_elems = buf.size() / elem_size;
 
-                auto output_begin = reinterpret_cast<rocfft_complex<_Float16>*>(buf.data());
+                auto output_begin = reinterpret_cast<rocfft_complex<rocfft_fp16>*>(buf.data());
                 for(size_t i = 0; i < num_elems; ++i)
                 {
                     auto& element = output_begin[i];
@@ -390,10 +668,10 @@ void apply_store_callback(const fft_params& params, std::vector<hostbuf>& output
         {
         case fft_precision_half:
         {
-            const size_t elem_size = sizeof(_Float16);
+            const size_t elem_size = sizeof(rocfft_fp16);
             const size_t num_elems = output.front().size() / elem_size;
 
-            auto output_begin = reinterpret_cast<_Float16*>(output.front().data());
+            auto output_begin = reinterpret_cast<rocfft_fp16*>(output.front().data());
             for(size_t i = 0; i < num_elems; ++i)
             {
                 auto& element = output_begin[i];
@@ -466,10 +744,10 @@ void apply_load_callback(const fft_params& params, std::vector<hostbuf>& input)
         {
         case fft_precision_half:
         {
-            const size_t elem_size = sizeof(std::complex<_Float16>);
+            const size_t elem_size = sizeof(std::complex<rocfft_fp16>);
             const size_t num_elems = input.front().size() / elem_size;
 
-            auto input_begin = reinterpret_cast<rocfft_complex<_Float16>*>(input.front().data());
+            auto input_begin = reinterpret_cast<rocfft_complex<rocfft_fp16>*>(input.front().data());
             for(size_t i = 0; i < num_elems; ++i)
             {
                 input_begin[i] = load_callback(input_begin, i, &cbdata, nullptr);
@@ -509,10 +787,10 @@ void apply_load_callback(const fft_params& params, std::vector<hostbuf>& input)
         {
         case fft_precision_half:
         {
-            const size_t elem_size = sizeof(_Float16);
+            const size_t elem_size = sizeof(rocfft_fp16);
             const size_t num_elems = input.front().size() / elem_size;
 
-            auto input_begin = reinterpret_cast<_Float16*>(input.front().data());
+            auto input_begin = reinterpret_cast<rocfft_fp16*>(input.front().data());
             for(size_t i = 0; i < num_elems; ++i)
             {
                 input_begin[i] = load_callback(input_begin, i, &cbdata, nullptr);
@@ -573,11 +851,11 @@ void apply_store_callback(const fft_params& params, std::vector<hostbuf>& output
     {
     case fft_precision_half:
     {
-        const size_t elem_size = sizeof(_Float16);
+        const size_t elem_size = sizeof(rocfft_fp16);
         for(auto& buf : output)
         {
             const size_t num_elems    = buf.size() / elem_size;
-            auto         output_begin = reinterpret_cast<_Float16*>(buf.data());
+            auto         output_begin = reinterpret_cast<rocfft_fp16*>(buf.data());
             for(size_t i = 0; i < num_elems; ++i)
             {
                 auto& element = output_begin[i];
