@@ -36,6 +36,11 @@ const static std::vector<fft_precision> precision_range_sp_dp
 
 const static std::vector<fft_result_placement> place_range
     = {fft_placement_inplace, fft_placement_notinplace};
+const static std::vector<fft_transform_type> trans_type_range_full
+    = {fft_transform_type_complex_forward,
+       fft_transform_type_real_forward,
+       fft_transform_type_complex_inverse,
+       fft_transform_type_real_inverse};
 const static std::vector<fft_transform_type> trans_type_range
     = {fft_transform_type_complex_forward, fft_transform_type_real_forward};
 const static std::vector<fft_transform_type> trans_type_range_complex
@@ -51,13 +56,34 @@ inline double hash_prob(const int seed, const std::string& token)
     // allows one to run the same tests for a given
     // random seed; ie the test suite is repeatable.
     std::hash<std::string>           hasher;
-    std::ranlux24_base               gen(random_seed + hasher(token));
+    std::ranlux24_base               gen(seed + hasher(token));
     std::uniform_real_distribution<> dis(0.0, 1.0);
 
     const double roll = dis(gen);
     return roll;
 }
 
+template <typename T, std::enable_if_t<std::is_arithmetic_v<T>, bool> = true>
+std::vector<T> merge_and_sort_values(const std::vector<std::vector<T>>& set_of_vecs,
+                                     size_t max_num_elem = std::numeric_limits<size_t>::max())
+{
+    std::vector<T> merged;
+    for(const auto& vec : set_of_vecs)
+    {
+        std::copy(vec.begin(), vec.end(), std::back_inserter(merged));
+    }
+    std::sort(merged.begin(), merged.end());
+    auto last_unique = std::unique(merged.begin(), merged.end());
+    merged.erase(last_unique, merged.end());
+    std::ranlux24_base gen(random_seed);
+    while(merged.size() > max_num_elem)
+    {
+        // remove pseudo-randomly chosen elements
+        merged.erase(merged.begin() + (static_cast<size_t>(gen()) % merged.size()));
+    }
+    return merged;
+}
+
 // Given a vector of vector of lengths, generate all unique permutations.
 // Add an optional vector of ad-hoc lengths to the result.
 inline std::vector<std::vector<size_t>>
@@ -236,9 +262,10 @@ struct stride_generator_3D_inner_batch : public stride_generator
 
 // Create an array of parameters to pass to gtest.  Base generator
 // that allows choosing transform type.
-inline auto param_generator_base(const std::vector<fft_transform_type>&   type_range,
+inline auto param_generator_base(const double                             base_prob,
+                                 const std::vector<fft_transform_type>&   type_range,
                                  const std::vector<std::vector<size_t>>&  v_lengths,
-                                 const std::vector<fft_precision>&        precision_range,
+                                 const std::vector<fft_precision>&        precisions,
                                  const std::vector<size_t>&               batch_range,
                                  decltype(generate_types)                 types_generator,
                                  const stride_generator&                  istride,
@@ -247,11 +274,14 @@ inline auto param_generator_base(const std::vector<fft_transform_type>&   type_r
                                  const std::vector<std::vector<size_t>>&  ooffset_range,
                                  const std::vector<fft_result_placement>& place_range,
                                  const bool                               planar        = true,
-                                 const bool                               run_callbacks = false)
+                                 const bool                               run_callbacks = false,
+                                 const fft_auto_allocation auto_alloc = fft_auto_allocation_default)
 {
-
     std::vector<fft_params> params;
 
+    if(base_prob == 0)
+        return params;
+
     // For any length, we compute double-precision CPU reference
     // for largest batch size first and reuse for smaller batch
     // sizes, then convert to single-precision.
@@ -267,7 +297,7 @@ inline auto param_generator_base(const std::vector<fft_transform_type>&   type_r
                 continue;
             }
             {
-                for(const auto precision : precision_range)
+                for(const auto precision : precisions)
                 {
                     for(const auto batch : batch_range)
                     {
@@ -297,6 +327,7 @@ inline auto param_generator_base(const std::vector<fft_transform_type>&   type_r
                                             param.otype          = std::get<3>(types);
                                             param.ioffset        = ioffset;
                                             param.ooffset        = ooffset;
+                                            param.auto_allocate  = auto_alloc;
 
                                             if(run_callbacks)
                                             {
@@ -319,16 +350,21 @@ inline auto param_generator_base(const std::vector<fft_transform_type>&   type_r
                                             const double roll
                                                 = hash_prob(random_seed, param.token());
                                             const double run_prob
-                                                = test_prob
-                                                  * (param.is_planar() ? planar_prob : 1.0)
-                                                  * (run_callbacks ? callback_prob : 1.0);
+                                                = base_prob
+                                                  * (param.is_planar() ? complex_planar_prob_factor
+                                                                       : 1.0)
+                                                  * (param.is_interleaved()
+                                                         ? complex_interleaved_prob_factor
+                                                         : 1.0)
+                                                  * (param.is_real() ? real_prob_factor : 1.0)
+                                                  * (run_callbacks ? callback_prob_factor : 1.0);
 
                                             if(roll > run_prob)
                                             {
                                                 if(verbose > 4)
                                                 {
-                                                    std::cout << "Test skipped (probability "
-                                                              << run_prob << " > " << roll << ")\n";
+                                                    std::cout << "Test skipped: (roll=" << roll
+                                                              << " > " << run_prob << ")\n";
                                                 }
                                                 continue;
                                             }
@@ -351,7 +387,8 @@ inline auto param_generator_base(const std::vector<fft_transform_type>&   type_r
 
 // Create an array of parameters to pass to gtest.  Default generator
 // that picks all transform types.
-inline auto param_generator(const std::vector<std::vector<size_t>>&  v_lengths,
+inline auto param_generator(const double                             base_prob,
+                            const std::vector<std::vector<size_t>>&  v_lengths,
                             const std::vector<fft_precision>&        precision_range,
                             const std::vector<size_t>&               batch_range,
                             const stride_generator&                  istride,
@@ -360,9 +397,11 @@ inline auto param_generator(const std::vector<std::vector<size_t>>&  v_lengths,
                             const std::vector<std::vector<size_t>>&  ooffset_range,
                             const std::vector<fft_result_placement>& place_range,
                             const bool                               planar,
-                            const bool                               run_callbacks = false)
+                            const bool                               run_callbacks = false,
+                            const fft_auto_allocation auto_alloc = fft_auto_allocation_default)
 {
-    return param_generator_base(trans_type_range,
+    return param_generator_base(base_prob,
+                                trans_type_range,
                                 v_lengths,
                                 precision_range,
                                 batch_range,
@@ -373,11 +412,13 @@ inline auto param_generator(const std::vector<std::vector<size_t>>&  v_lengths,
                                 ooffset_range,
                                 place_range,
                                 planar,
-                                run_callbacks);
+                                run_callbacks,
+                                auto_alloc);
 }
 
 // Create an array of parameters to pass to gtest.  Only tests complex-type transforms
-inline auto param_generator_complex(const std::vector<std::vector<size_t>>&  v_lengths,
+inline auto param_generator_complex(const double                             base_prob,
+                                    const std::vector<std::vector<size_t>>&  v_lengths,
                                     const std::vector<fft_precision>&        precision_range,
                                     const std::vector<size_t>&               batch_range,
                                     const stride_generator&                  istride,
@@ -386,9 +427,12 @@ inline auto param_generator_complex(const std::vector<std::vector<size_t>>&  v_l
                                     const std::vector<std::vector<size_t>>&  ooffset_range,
                                     const std::vector<fft_result_placement>& place_range,
                                     const bool                               planar,
-                                    const bool                               run_callbacks = false)
+                                    const bool                               run_callbacks = false,
+                                    const fft_auto_allocation                auto_alloc
+                                    = fft_auto_allocation_default)
 {
-    return param_generator_base(trans_type_range_complex,
+    return param_generator_base(base_prob,
+                                trans_type_range_complex,
                                 v_lengths,
                                 precision_range,
                                 batch_range,
@@ -399,11 +443,13 @@ inline auto param_generator_complex(const std::vector<std::vector<size_t>>&  v_l
                                 ooffset_range,
                                 place_range,
                                 planar,
-                                run_callbacks);
+                                run_callbacks,
+                                auto_alloc);
 }
 
 // Create an array of parameters to pass to gtest.
-inline auto param_generator_real(const std::vector<std::vector<size_t>>&  v_lengths,
+inline auto param_generator_real(const double                             base_prob,
+                                 const std::vector<std::vector<size_t>>&  v_lengths,
                                  const std::vector<fft_precision>&        precision_range,
                                  const std::vector<size_t>&               batch_range,
                                  const stride_generator&                  istride,
@@ -412,9 +458,11 @@ inline auto param_generator_real(const std::vector<std::vector<size_t>>&  v_leng
                                  const std::vector<std::vector<size_t>>&  ooffset_range,
                                  const std::vector<fft_result_placement>& place_range,
                                  const bool                               planar,
-                                 const bool                               run_callbacks = false)
+                                 const bool                               run_callbacks = false,
+                                 const fft_auto_allocation auto_alloc = fft_auto_allocation_default)
 {
-    return param_generator_base(trans_type_range_real,
+    return param_generator_base(base_prob,
+                                trans_type_range_real,
                                 v_lengths,
                                 precision_range,
                                 batch_range,
@@ -425,18 +473,47 @@ inline auto param_generator_real(const std::vector<std::vector<size_t>>&  v_leng
                                 ooffset_range,
                                 place_range,
                                 planar,
-                                run_callbacks);
+                                run_callbacks,
+                                auto_alloc);
 }
 
 template <class Tcontainer>
-auto param_generator_token(const Tcontainer& tokens)
+auto param_generator_token(const double base_prob, const Tcontainer& tokens)
 {
     std::vector<fft_params> params;
-    params.reserve(tokens.size());
-    for(auto t : tokens)
+
+    if(base_prob == 0)
+        return params;
+
+    for(const auto& token : tokens)
     {
-        params.push_back({});
-        params.back().from_token(t);
+        fft_params param;
+        param.from_token(token);
+        param.validate();
+        if(param.valid())
+        {
+            const double roll = hash_prob(random_seed, param.token());
+            const double run_prob
+                = base_prob * (param.is_planar() ? complex_planar_prob_factor : 1.0)
+                  * (param.is_interleaved() ? complex_interleaved_prob_factor : 1.0)
+                  * (param.is_real() ? real_prob_factor : 1.0)
+                  * (param.is_callback() ? callback_prob_factor : 1.0);
+
+            if(roll > run_prob)
+            {
+                if(verbose > 4)
+                {
+                    std::cout << "Test skipped: (roll=" << roll << " > " << run_prob << ")\n";
+                }
+                continue;
+            }
+
+            params.push_back(param);
+        }
+        else
+        {
+            std::cout << "invalid token\n";
+        }
     }
     return params;
 }
